#pragma once

#include <chrono>
#include <rewindow/types/key.h>
#include <rewindow/types/mousebutton.h>

namespace ReWindow {

    using namespace std::chrono_literals;
    using precision_clock = std::chrono::high_resolution_clock;
    using precision_timestamp = precision_clock::time_point;

    class TimestampedEvent {
    private:

    public:
        precision_timestamp Timestamp;
        TimestampedEvent() : Timestamp(precision_clock::now())
        { }
    };

    class RWindowEvent : public TimestampedEvent {
    public:
        RWindowEvent() : TimestampedEvent() { }
    };

    class KeyboardEvent : public RWindowEvent {
    public:
        Key key;
        KeyState state;
        KeyboardEvent(Key key, KeyState state) : RWindowEvent(), key(key), state(state) {}
    };
    class MouseEvent : public RWindowEvent {};

    class GamepadEvent : public RWindowEvent {};

    class MouseMoveEvent : public MouseEvent {
    public:
        Vector2 Position;
        Vector2 Delta;
        MouseMoveEvent(const Vector2 &pos) : MouseEvent(), Position(pos)
        {}
        MouseMoveEvent(int x, int y) : MouseEvent(), Position(Vector2(x, y))
        {}
    };

    class KeyDownEvent : public KeyboardEvent {
    public:
        KeyDownEvent(Key key) : KeyboardEvent(key, KeyState::Pressed) {}
    };

    class KeyUpEvent : public KeyboardEvent {
    public:
        KeyUpEvent(Key key) : KeyboardEvent(key, KeyState::Released) {}
    };

    class MouseWheelEvent : public MouseEvent
    {
    public:
        int WheelMovement;
        MouseWheelEvent() = default;
        MouseWheelEvent(int wheel) : MouseEvent(), WheelMovement(wheel) {}
    };

    class MouseButtonDownEvent : public MouseEvent {
    public:
        MouseButton Button;

        MouseButtonDownEvent() = default;
        MouseButtonDownEvent(MouseButton button) : MouseEvent(), Button(button) {}
    };

    class MouseButtonUpEvent : public MouseEvent {
    public:
        MouseButton Button;
        MouseButtonUpEvent() = default;
        MouseButtonUpEvent(MouseButton button) : MouseEvent(), Button(button) {}
    };

    class WindowResizeRequestEvent : public RWindowEvent
    {
    public:
        Vector2 Size;
    };
}