#pragma once

#if __linux__
#include <X11/cursorfont.h>
#endif

namespace ReWindow
{
#if __linux__
	enum class X11CursorStyle
	{
		Default = XC_left_ptr,
		X = XC_X_cursor,
		Arrow = XC_arrow,
		IBeam = XC_xterm,
		BottomLeftCorner = XC_bottom_left_corner,
		BottomRightCorner = XC_bottom_right_corner,
		BottomSide = XC_bottom_side,
		Dot = XC_dot,
		DoubleArrow = XC_double_arrow,
		Exchange = XC_exchange,
		Hand = XC_hand2,
		LeftSide = XC_left_side,
		Plus = XC_plus,
		RightSide = XC_right_side,
		Pencil = XC_pencil
	};

	class CursorStyle {
		public:
			X11CursorStyle X11Cursor;
			CursorStyle(X11CursorStyle style) : X11Cursor(style) {}
	};

	namespace Cursors {
		static const CursorStyle Default {X11CursorStyle::Default};
		static const CursorStyle X {X11CursorStyle::X};
		static const CursorStyle Arrow {X11CursorStyle::Arrow};
		static const CursorStyle IBeam {X11CursorStyle::IBeam};
		static const CursorStyle BottomLeftCorner {X11CursorStyle::BottomLeftCorner};
		static const CursorStyle BottomRightCorner {X11CursorStyle::BottomRightCorner};
		static const CursorStyle BottomSide {X11CursorStyle::BottomSide};
		static const CursorStyle Dot {X11CursorStyle::Dot};
		static const CursorStyle DoubleArrow {X11CursorStyle::DoubleArrow};
		static const CursorStyle Exchange {X11CursorStyle::Exchange};
		static const CursorStyle Hand {X11CursorStyle::Hand};
		static const CursorStyle LeftSide {X11CursorStyle::LeftSide};
		static const CursorStyle Plus {X11CursorStyle::Plus};
		static const CursorStyle RightSide {X11CursorStyle::RightSide};
		static const CursorStyle Pencil {X11CursorStyle::Pencil};
	}

#else
	// https://learn.microsoft.com/en-us/windows/win32/menurc/about-cursors
	enum WindowsCursorStyle {
		Arrow,
		IBeam,
		Wait,
		Cross,
		Hand,
		AppStarting,
	};

	class CursorStyle {
		public:
			WindowsCursorStyle WindowsCursor;
			CursorStyle (WindowsCursorStyle style): WindowsCursor(style) {}
	};

	namespace Cursors {
		static const CursorStyle Default {WindowsCursorStyle::Arrow};
		static const CursorStyle Arrow {WindowsCursorStyle::Arrow};
		static const CursorStyle IBeam {WindowsCursorStyle::IBeam};
		static const CursorStyle Cross {WindowsCursorStyle::Cross};
		static const CursorStyle Wait {WindowsCursorStyle::Wait};
		static const CursorStyle Hand {WindowsCursorStyle::Hand};
		static const CursorStyle AppStarting {WindowsCursorStyle::AppStarting};
	}
#endif
}