//
// ~DAWSH
#pragma once

#include <vector>
#include <stdexcept>
#include <iostream>
#include <string>
#include <rewindow/data/X11Scancodes.h>
#include <rewindow/data/WindowsScancodes.h>
#include <J3ML/LinearAlgebra.hpp>


class Key
{
private:
    //CLion on Linux falsely marks this as being wrong.
    static inline std::vector<Key> keyboard{};

public:
    static std::vector<Key> GetKeyboard();
    Key() = default;
    Key(const char* charcode, X11Scancode scancode, WindowsScancode wSc);
    std::string Mnemonic;
    X11Scancode x11ScanCode;
    WindowsScancode winScanCode;
    bool operator==(const Key& rhs) const;
    bool operator<(const Key& rhs) const;
    Key(const Key&) = default;
};

namespace Keys {
    // TODO: Encode both Uppercase and Lowercase version for each keymap
    static const Key Escape {"ESC", X11Scancode::ESCAPE, WindowsScancode::ESCAPE};
    static const Key F1  {"F1", X11Scancode::F1, WindowsScancode::F1};
    static const Key F2  {"F2", X11Scancode::F2, WindowsScancode::F2};
    static const Key F3  {"F3", X11Scancode::F3, WindowsScancode::F3};
    static const Key F4  {"F4", X11Scancode::F4, WindowsScancode::F4};
    static const Key F5  {"F5", X11Scancode::F5, WindowsScancode::F5};
    static const Key F6  {"F6", X11Scancode::F6, WindowsScancode::F6};
    static const Key F7  {"F7", X11Scancode::F7, WindowsScancode::F7};
    static const Key F8  {"F8", X11Scancode::F8, WindowsScancode::F8};
    static const Key F9  {"F9", X11Scancode::F9, WindowsScancode::F9};
    static const Key F10 {"F10", X11Scancode::F10, WindowsScancode::F10};
    static const Key F11 {"F11", X11Scancode::F11, WindowsScancode::F11};
    static const Key F12 {"F12", X11Scancode::F12, WindowsScancode::F12};
    static const Key Print {"PRINT", X11Scancode::PRINT, WindowsScancode::PRINTSCREEN};
    static const Key ScrollLock {"SCROLL_LOCK", X11Scancode::SCROLL_LOCK, WindowsScancode::SCROLL_LOCK};
    static const Key Break {"BREAK", X11Scancode::BREAK, WindowsScancode::BREAK};

    //TODO On Windows, Return & KP_RETURN are the same thing.
    static const Key NumPadReturn {"KP_↵", X11Scancode::KP_RETURN, WindowsScancode::NUMPAD_ENTER};
    static const Key NumPadPlus {"KP_+", X11Scancode::KP_PLUS, WindowsScancode::NUMPAD_PLUS};
    static const Key NumPadMinus {"KP_-", X11Scancode::KP_MINUS, WindowsScancode::NUMPAD_MINUS};
    static const Key NumPadAsterisk {"KP_*", X11Scancode::KP_ASTERISK, WindowsScancode::NUMPAD_ASTERISK};
    static const Key NumPadForwardSlash {"KP_/", X11Scancode::KP_SLASH, WindowsScancode::NUMPAD_SLASH};
    static const Key NumPadPeriod {"KP_.", X11Scancode::KP_PERIOD, WindowsScancode::NUMPAD_PERIOD};
    static const Key NumPadNumLock {"KP_NUMLOCK", X11Scancode::KP_NUMLOCK, WindowsScancode::NUM_LOCK};

    static const Key NumPad1 {"KP_1", X11Scancode::KP_1, WindowsScancode::NUMPAD_ONE};
    static const Key NumPad2 {"KP_2", X11Scancode::KP_2, WindowsScancode::NUMPAD_TWO};
    static const Key NumPad3 {"KP_3", X11Scancode::KP_3, WindowsScancode::NUMPAD_THREE};
    static const Key NumPad4 {"KP_4", X11Scancode::KP_4, WindowsScancode::NUMPAD_FOUR};
    static const Key NumPad5 {"KP_5", X11Scancode::KP_5, WindowsScancode::NUMPAD_FIVE};
    static const Key NumPad6 {"KP_6", X11Scancode::KP_6, WindowsScancode::NUMPAD_SIX};
    static const Key NumPad7 {"KP_7", X11Scancode::KP_7, WindowsScancode::NUMPAD_SEVEN};
    static const Key NumPad8 {"kP_8", X11Scancode::KP_8, WindowsScancode::NUMPAD_EIGHT};
    static const Key NumPad9 {"KP_9", X11Scancode::KP_9, WindowsScancode::NUMPAD_NINE};
    static const Key NumPad0 {"KP_0", X11Scancode::KP_0, WindowsScancode::NUMPAD_ZERO};

    static const Key Grave {"`", X11Scancode::GRAVE, WindowsScancode::GRAVE};
    static const Key One {"1", X11Scancode::ONE, WindowsScancode::ONE};
    static const Key Two {"2", X11Scancode::TWO, WindowsScancode::TWO};
    static const Key Three {"3", X11Scancode::THREE, WindowsScancode::THREE};
    static const Key Four {"4", X11Scancode::FOUR, WindowsScancode::FOUR};
    static const Key Five {"5", X11Scancode::FIVE, WindowsScancode::FIVE};
    static const Key Six {"6", X11Scancode::SIX, WindowsScancode::SIX};
    static const Key Seven {"7", X11Scancode::SEVEN, WindowsScancode::SEVEN};
    static const Key Eight {"8", X11Scancode::EIGHT, WindowsScancode::EIGHT};
    static const Key Nine {"9", X11Scancode::NINE, WindowsScancode::NINE};
    static const Key Zero {"0", X11Scancode::ZERO, WindowsScancode::ZERO};
    static const Key Minus {"-", X11Scancode::MINUS, WindowsScancode::MINUS};
    static const Key Equals {"+", X11Scancode::EQUALS, WindowsScancode::EQUALS};
    static const Key Backspace {"⌫", X11Scancode::BACKSPACE, WindowsScancode::BACKSPACE};

    static const Key Tab {"⇥", X11Scancode::TAB, WindowsScancode::TAB};
    static const Key Q {"Q", X11Scancode::Q, WindowsScancode::Q};
    static const Key W {"W", X11Scancode::W, WindowsScancode::W};
    static const Key E {"E", X11Scancode::E, WindowsScancode::E};
    static const Key R {"R", X11Scancode::R, WindowsScancode::R};
    static const Key T {"T", X11Scancode::T, WindowsScancode::T};
    static const Key Y {"Y", X11Scancode::Y, WindowsScancode::Y};
    static const Key U {"U", X11Scancode::U, WindowsScancode::U};
    static const Key I {"I", X11Scancode::I, WindowsScancode::I};
    static const Key O {"O", X11Scancode::O, WindowsScancode::O};
    static const Key P {"P", X11Scancode::P, WindowsScancode::P};
    static const Key LeftBracket {"[", X11Scancode::OPENING_SQUARE_BRACKET, WindowsScancode::LEFT_BRACKET};
    static const Key RightBracket {"]", X11Scancode::CLOSING_SQUARE_BRACKET, WindowsScancode::RIGHT_BRACKET};
    static const Key BackSlash {"\\", X11Scancode::BACKSLASH, WindowsScancode::BACKSLASH};

    static const Key CapsLock = {"CAPS", X11Scancode::CAPSLOCK, WindowsScancode::CAPS_LOCK};
    static const Key A {"A", X11Scancode::A, WindowsScancode::A};
    static const Key S {"S", X11Scancode::S, WindowsScancode::S};
    static const Key D {"D", X11Scancode::D, WindowsScancode::D};
    static const Key F {"F", X11Scancode::F, WindowsScancode::F};
    static const Key G {"G", X11Scancode::G, WindowsScancode::G};
    static const Key H {"H", X11Scancode::H, WindowsScancode::H};
    static const Key J {"J", X11Scancode::J, WindowsScancode::J};
    static const Key K {"K", X11Scancode::K, WindowsScancode::K};
    static const Key L {"L", X11Scancode::L, WindowsScancode::L};
    static const Key Semicolon {";", X11Scancode::SEMICOLON, WindowsScancode::SEMICOLON};
    static const Key SingeQuote {"\'", X11Scancode::SINGLEQUOTE, WindowsScancode::SINGLEQUOTE};
    static const Key Return {"↵", X11Scancode::RETURN, WindowsScancode::ENTER};

    static const Key LeftShift {"⇧", X11Scancode::LEFT_SHIFT, WindowsScancode::LEFT_SHIFT};
    static const Key Z {"Z", X11Scancode::Z, WindowsScancode::Z};
    static const Key X {"X", X11Scancode::X, WindowsScancode::X};
    static const Key C {"C", X11Scancode::C, WindowsScancode::C};
    static const Key V {"V", X11Scancode::V, WindowsScancode::V};
    static const Key B {"B", X11Scancode::B, WindowsScancode::B};
    static const Key N {"N", X11Scancode::N, WindowsScancode::N};
    static const Key M {"M", X11Scancode::M, WindowsScancode::M};
    static const Key Comma = {",", X11Scancode::COMMA, WindowsScancode::COMMA};
    static const Key Period {".", X11Scancode::PERIOD, WindowsScancode::PERIOD};
    static const Key ForwardSlash {"/", X11Scancode::SLASH, WindowsScancode::SLASH};

    static const Key LeftControl {"LCTRL", X11Scancode::LEFT_CTRL, WindowsScancode::CTRL};
    static const Key Super {"❖", X11Scancode::SUPER, WindowsScancode::SUPER};
    static const Key LeftAlt {"🄰", X11Scancode::LEFT_ALT, WindowsScancode::ALT};
    static const Key Space {" ", X11Scancode::SPACE, WindowsScancode::SPACE};
    static const Key RightAlt {"R🄰", X11Scancode::RIGHT_ALT, WindowsScancode::ALT};
    static const Key Menu = {"▤", X11Scancode::MENU, WindowsScancode::MENU};
    static const Key RightControl {"RCTRL", X11Scancode::RIGHT_CONTROL, WindowsScancode::CTRL};
    static const Key RightShift {"R⇧", X11Scancode::RIGHT_SHIFT, WindowsScancode::RIGHT_SHIFT};

    static const Key Insert {"INSERT", X11Scancode::INSERT, WindowsScancode::INSERT};
    static const Key Home {"HOME", X11Scancode::HOME, WindowsScancode::HOME};
    static const Key PageUp {"PAGEUP", X11Scancode::PAGEUP, WindowsScancode::PAGE_UP};
    static const Key Delete {"DELETE", X11Scancode::DEL, WindowsScancode::DEL};
    static const Key End {"End", X11Scancode::END, WindowsScancode::END};
    static const Key PageDown {"PAGEDOWN", X11Scancode::PAGEDOWN, WindowsScancode::PAGE_DOWN};

    static const Key UpArrow {reinterpret_cast<const char*>(u8"↑"), X11Scancode::UP, WindowsScancode::UP_ARROW};
    static const Key DownArrow {reinterpret_cast<const char*>(u8"↓"), X11Scancode::DOWN, WindowsScancode::DOWN_ARROW};
    static const Key LeftArrow {reinterpret_cast<const char*>(u8"←"), X11Scancode::LEFT, WindowsScancode::LEFT_ARROW};
    static const Key RightArrow {reinterpret_cast<const char*>(u8"→"), X11Scancode::RIGHT, WindowsScancode::RIGHT_ARROW};

}

enum class KeyState {Pressed, Released};



Key GetKeyFromX11Scancode(X11Scancode code);
Key GetKeyFromWindowsScancode(WindowsScancode code);