/// ReWindowLibrary
/// A C++20 Library for creating and managing windows in a platform-independent manner
/// Developed and Maintained by the boys @ Redacted Software.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file mousebutton.hpp
/// @desc MouseButton class and defined MouseButtons.
/// @edit 2024-07-29

#pragma once

#include <string>

class MouseButton {
public:
    std::string Mnemonic;
    unsigned int ButtonIndex;

    MouseButton() = default;
    explicit MouseButton(const std::string& charcode, unsigned int index);

    bool operator == (const MouseButton& mb) const;
    bool operator<(const MouseButton& rhs) const;
    MouseButton(const MouseButton&) = default;
};


namespace MouseButtons
{
    static const MouseButton Left           ("L", 1);
    static const MouseButton Right          ("R", 2);
    static const MouseButton Middle         ("M", 3);
    static const MouseButton Mouse4         ("4", 8);
    static const MouseButton Mouse5         ("5", 9);
    static const MouseButton Unimplemented  ("?", 0);

    /// NOTE: IsMouseButtonDown will not return correctly for the mouse-wheel-buttons, because the action is effectively instantaneous.
    static const MouseButton MWheelUp       ("U", 4);
    /// NOTE: IsMouseButtonDown will not return correctly for the mouse-wheel-buttons, because the action is effectively instantaneous.
    static const MouseButton MWheelDown     ("D", 5);
}

MouseButton GetMouseButtonFromXButton(unsigned int button);


