#include <rewindow/types/display.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/extensions/Xrandr.h>
#include <stdexcept>

using namespace ReWindow;

std::vector<ReWindow::Display> ReWindow::Display::getDisplaysFromWindowManager() {
    std::vector<ReWindow::Display> result;
    _XDisplay* display = XOpenDisplay(nullptr);

    if (!display)
        throw std::runtime_error("There isn't an X11 root display????");

    Window root_window = DefaultRootWindow(display);
    XRRScreenResources* screen_resources = XRRGetScreenResources(display, root_window);

    if (!screen_resources)
        throw std::runtime_error("Failed to get screen resources.");

    for (int i = 0; i < screen_resources->noutput; i++) {
        RROutput output = screen_resources->outputs[i];
        XRROutputInfo* output_info = XRRGetOutputInfo(display, screen_resources, output);

        if (output_info->connection == RR_Connected) {
            ReWindow::Display d;
            XRRCrtcInfo* crtc_info = XRRGetCrtcInfo(display, screen_resources, output_info->crtc);

            if (!crtc_info)
                throw std::runtime_error("Failed to get CRT info.");

            u32 default_width = (u32) crtc_info->width;
            u32 default_height = (u32) crtc_info->height;
            short default_refresh_rate = 0;

            for (int k = 0; k < screen_resources->nmode; k++) {
                if (screen_resources->modes[k].id == crtc_info->mode) {
                    XRRModeInfo mode_info = screen_resources->modes[k];
                    double refresh_rate;

                    if (mode_info.hTotal > 0 && mode_info.vTotal > 0)
                        refresh_rate = mode_info.dotClock / (double)(mode_info.hTotal * mode_info.vTotal),
                        default_refresh_rate = (short)refresh_rate;

                    d.default_graphics_mode = {default_width, default_height, default_refresh_rate};
                    break;
                }
            }

            for (int j = 0; j < output_info->nmode; j++) {
                RRMode mode = output_info->modes[j];
                for (int k = 0; k < screen_resources->nmode; k++) {
                    if (screen_resources->modes[k].id == mode) {
                        XRRModeInfo mode_info = screen_resources->modes[k];
                        // Calculate Refresh rate from the mode's dot clock
                        double refresh_rate = 0.0;
                        if (mode_info.hTotal > 0 && mode_info.vTotal > 0)
                            refresh_rate = mode_info.dotClock / (double)(mode_info.hTotal * mode_info.vTotal);

                        d.graphics_modes.emplace_back((u32) mode_info.width,(u32) mode_info.height,(short) refresh_rate);
                    }
                }
            }
            XRRFreeCrtcInfo(crtc_info);
            result.push_back(d);
        }
        XRRFreeOutputInfo(output_info);
    }

    XRRFreeScreenResources(screen_resources);
    XCloseDisplay(display);
    return result;
}