#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <GL/glx.h>
#include <cstdlib>
#include <cstring>
#include <rewindow/types/window.h>
#include <rewindow/types/cursors.h>
#include <J3ML/J3ML.hpp>
#include <rewindow/logger/logger.h>



// TODO: Move all "global" members to be instantiated class members of Window
// Doing this would break the intended "Platform-Specific" Encapsulation
// So should we do derived platform-specific subclasses?
// The intended goal of the ReWindow class is a one-stop object that handles window management on **ALL** platforms

using namespace ReWindow;

/*
Window window;
XEvent xev;
Display* display = XOpenDisplay(nullptr);
int defaultScreen = DefaultScreen(display);
XVisualInfo* visual;
XSetWindowAttributes xSetWindowAttributes;
XWindowAttributes windowAttributes;
Atom wmDeleteWindow;
// Make it start as floating because fuck tiling WMs
Atom windowTypeAtom;
Atom windowTypeUtilityAtom;
XSizeHints hints;
GLXContext glContext;
Cursor invisible_cursor = 0;

Vector2 render_area_position = {0, 0};
Vector2 position = {0, 0};
bool should_poll_x_for_mouse_pos = true;
 */

class RWindowImpl::Platform {
public:
    Platform() = default;
public:
    Window window;
    XEvent xev;
    Display* display = XOpenDisplay(nullptr);
    int defaultScreen = DefaultScreen(display);
    XVisualInfo* visual;
    XSetWindowAttributes xSetWindowAttributes;
    XWindowAttributes windowAttributes;
    Atom wmDeleteWindow;
    // Make it start as floating because fuck tiling WMs
    Atom windowTypeAtom;
    Atom windowTypeUtilityAtom;
    XSizeHints hints;
    GLXContext glContext;
    Cursor invisible_cursor = 0;

    Vector2 render_area_position = {0, 0};
    Vector2 position = {0, 0};
    bool should_poll_x_for_mouse_pos = true;
};

RWindowImpl::RWindowImpl(const std::string &wTitle, int wWidth, int wHeight, RenderingAPI wRenderer, bool wFullscreen, bool wResizable, bool wVsync) : pPlatform(new Platform) {
    title = wTitle;
    width = wWidth;
    height = wHeight;
    renderer = wRenderer;
    fullscreen_mode = wFullscreen;
    resizable = wResizable;
    vsync = wVsync;
}

RWindowImpl::~RWindowImpl() {
    if (pPlatform) { std::cout << "FUCK" << std::endl; delete pPlatform; };
    if (open)
        DestroyOSWindowHandle();
}

//using namespace ReWindow;

void RWindowImpl::Raise() const {
    Logger::Debug(std::format("Raising window '{}'", this->title));

    // Get the position of the renderable area relative to the rest of the window.
    XGetWindowAttributes(pPlatform->display, pPlatform->window, &pPlatform->windowAttributes);
    pPlatform->render_area_position =  { (float) pPlatform->windowAttributes.x, (float) pPlatform->windowAttributes.y };

    XRaiseWindow(pPlatform->display, pPlatform->window);
}

void RWindowImpl::Lower() const
{
    Logger::Debug(std::format("Lowering window '{}'", this->title));
    XLowerWindow(pPlatform->display, pPlatform->window);
}

void RWindowImpl::DestroyOSWindowHandle() {
    Logger::Debug(std::format("Destroying window '{}'", this->title));
    XDestroySubwindows(pPlatform->display, pPlatform->window);
    Logger::Debug(std::format("Destroyed window '{}'", this->title));
    XDestroyWindow(pPlatform->display, pPlatform->window);

    system("xset r on"); // Re-set X-server parameter to enable key-repeat.

    open = false;
}

//void RWindow::

void RWindowImpl::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
    if (pPlatform->invisible_cursor == 0) {
        Pixmap blank_pixmap = XCreatePixmap(pPlatform->display, pPlatform->window, 1, 1, 1);
        XColor dummy; dummy.pixel = 0; dummy.red = 0; dummy.flags = 0;
        pPlatform->invisible_cursor = XCreatePixmapCursor(pPlatform->display, blank_pixmap, blank_pixmap, &dummy, &dummy, 0, 0);
        XFreePixmap(pPlatform->display, blank_pixmap);
    }

    if (!cursor_enable)
        XDefineCursor(pPlatform->display, pPlatform->window, pPlatform->invisible_cursor);

    if (cursor_enable)
        XUndefineCursor(pPlatform->display, pPlatform->window);
}

void RWindowImpl::SetResizable(bool sizable) {
    XGetWindowAttributes(pPlatform->display,pPlatform->window,&pPlatform->windowAttributes);


    this->resizable = sizable;
    if (!sizable)
    {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        pPlatform->hints.flags = PMinSize | PMaxSize;
        pPlatform->hints.min_width = pPlatform->hints.max_width = pPlatform->windowAttributes.width;
        pPlatform->hints.min_height = pPlatform->hints.max_height = pPlatform->windowAttributes.height;
        XSetWMNormalHints(pPlatform->display, pPlatform->window, &pPlatform->hints);
    }
    //this->SetFlag(RWindowFlags::RESIZABLE, resizable);

}

// Fuck you
void RWindowImpl::SetFlag(RWindowFlags flag, bool state) {
    XGetWindowAttributes(pPlatform->display,pPlatform->window,&pPlatform->windowAttributes);
    flags[(int) flag] = state;
    //Once you've done this you cannot make it resizable again.
    if (flag == RWindowFlags::RESIZABLE && !state) {
        Logger::Debug("Once you've done this you cannot make it resizable again.");
        pPlatform->hints.flags = PMinSize | PMaxSize;
        pPlatform->hints.min_width = pPlatform->hints.max_width = pPlatform->windowAttributes.width;
        pPlatform->hints.min_height = pPlatform->hints.max_height = pPlatform->windowAttributes.height;
        XSetWMNormalHints(pPlatform->display, pPlatform->window, &pPlatform->hints);
    }
    Logger::Debug(std::format("Set flag '{}' to state '{}' for window '{}'", RWindowFlagToStr(flag), state, this->title));
}

void RWindowImpl::PollEvents() {
    while(XPending(pPlatform->display)) {
        XNextEvent(pPlatform->display, &pPlatform->xev);


        if (pPlatform->xev.type == ClientMessage)
            Logger::Info(std::format("Event '{}'", "ClientMessage"));

        if (pPlatform->xev.xclient.message_type == XInternAtom(pPlatform->display, "WM_PROTOCOLS", False) && static_cast<Atom>(pPlatform->xev.xclient.data.l[0]) == pPlatform->wmDeleteWindow) {
         Logger::Info("Closing");
            Close();
        }

        if (pPlatform->xev.type == FocusIn) {
            Logger::Debug(std::format("Event'{}'", "FocusIn"));
            XAutoRepeatOff(pPlatform->display);
            SetFlag(RWindowFlags::IN_FOCUS, true);

            if (!cursor_visible)
                XDefineCursor(pPlatform->display, pPlatform->window, pPlatform->invisible_cursor);

            // Get the position of the renderable area relative to the rest of the window.
            XGetWindowAttributes(pPlatform->display, pPlatform->window, &pPlatform->windowAttributes);
            pPlatform->render_area_position =  { (float) pPlatform->windowAttributes.x, (float) pPlatform->windowAttributes.y };
            processFocusIn();
        }

        if (pPlatform->xev.type == FocusOut) {
            Logger::Debug(std::format("Event '{}'", "FocusOut"));
            XAutoRepeatOn(pPlatform->display);

            SetFlag(RWindowFlags::IN_FOCUS, false);
            if (!cursor_visible)
                XUndefineCursor(pPlatform->display, pPlatform->window);

            processFocusOut();
        }

        if (pPlatform->xev.type == KeyRelease) {
            Logger::Debug(std::format("Event '{}'", "KeyRelease"));
            auto scancode = (X11Scancode) pPlatform->xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyRelease(key);
        }

        if (pPlatform->xev.type == KeyPress) {
            Logger::Debug(std::format("Event '{}'", "KeyPress"));
            auto scancode = (X11Scancode) pPlatform->xev.xkey.keycode;
            auto key = GetKeyFromX11Scancode(scancode);
            processKeyPress(key);
        }

        if (pPlatform->xev.type == ButtonRelease) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (pPlatform->xev.xbutton.button != 4 && pPlatform->xev.xbutton.button != 5) {
                MouseButton button = GetMouseButtonFromXButton(pPlatform->xev.xbutton.button);
                Logger::Debug(std::format("Event '{}'", "ButtonRelease"));
                processMouseRelease(button);
            }
        }

        if (pPlatform->xev.type == ButtonPress) {

            // Mouse Wheel fires both the ButtonPress and ButtonRelease instantaneously.
            // Therefore, we handle it as a specific MouseWheel event rather than a MouseButton event,
            // and only call on ButtonPress, otherwise it will appear to duplicate the mouse wheel scroll.
            if (pPlatform->xev.xbutton.button == 4) {
                processMouseWheel(-1);
            } else if (pPlatform->xev.xbutton.button == 5) {
                processMouseWheel(1);
            } else
            {
                MouseButton button = GetMouseButtonFromXButton(pPlatform->xev.xbutton.button);

                Logger::Debug(std::format("Event: MouseButtonPress {}", button.Mnemonic));

                processMousePress(button);
            }
        }

        if (pPlatform->xev.type == Expose)
        {
            Logger::Debug(std::format("Event '{}'", "Expose"));
            // Essentially allows more than one window to be drawable
            // https://www.khronos.org/opengl/wiki/Programming_OpenGL_in_Linux:_GLX_and_Xlib
            // https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/glXMakeCurrent.xml
            //glXMakeCurrent(pPlatform->display, pPlatform->window, pPlatform->glContext);
        }

        // NOTE: This event is functionally useless, as it only informs of the very beginning and end of a mouse movement.
        if (pPlatform->xev.type == MotionNotify)
        {
            Logger::Debug(std::format("Event '{}'", "MotionNotify"));
        }

        if (pPlatform->xev.type == ConfigureNotify) {
            if (this->width != pPlatform->xev.xconfigurerequest.width || this->height != pPlatform->xev.xconfigurerequest.height) {
                Logger::Debug(std::format("Event '{}'", "ResizeRequest"));

                this->width = pPlatform->xev.xconfigurerequest.width;
                this->height = pPlatform->xev.xconfigurerequest.height;

                auto eventData = WindowResizeRequestEvent();
                eventData.Size = { (float) pPlatform->xev.xconfigurerequest.width, (float) pPlatform->xev.xconfigurerequest.height };
                lastKnownWindowSize = eventData.Size;

                OnResizeRequest(eventData);
                OnResizeRequestEvent(eventData);
            }

            //Window Moved.
            if (pPlatform->position.x != pPlatform->xev.xconfigurerequest.x || pPlatform->position.y != pPlatform->xev.xconfigurerequest.y)
                pPlatform->position = { (float) pPlatform->xev.xconfigurerequest.x, (float) pPlatform->xev.xconfigurerequest.y };
        }

    }
    previousKeyboard = currentKeyboard;
    previousMouse.Buttons = currentMouse.Buttons;
}


void RWindowImpl::Refresh() {
    // Essentially allows more than one window to be drawable
    // https://www.khronos.org/opengl/wiki/Programming_OpenGL_in_Linux:_GLX_and_Xlib
    // https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/glXMakeCurrent.xml
    glXMakeCurrent(pPlatform->display, pPlatform->window, pPlatform->glContext);

    PollEvents();
    OnRefresh(delta_time);

    // Only call once and cache the result.
    currentMouse.Position = GetAccurateMouseCoordinates();

    /// TODO: Implement optional minimum epsilon to trigger a Mouse Update.
    if (currentMouse.Position != previousMouse.Position) {
        processMouseMove(previousMouse.Position, currentMouse.Position);
        previousMouse.Position = currentMouse.Position;
    }
}


// Might make the window go off the screen on some window managers.
void RWindowImpl::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;

    this->width = newWidth;
    this->height = newHeight;
    XResizeWindow(pPlatform->display, pPlatform->window, newWidth, newHeight);
    XFlush(pPlatform->display);
    Logger::Info(std::format("Set size for '{}' to {} x {}", this->title, newWidth, newHeight));
}

Vector2 RWindowImpl::GetAccurateMouseCoordinates() const {

    Window root_return, child_return;
    int root_x_ret, root_y_ret;
    int win_x_ret, win_y_ret;
    uint32_t mask_return;

    // This seems to be relative to the top left corner of the renderable area.
    bool mouseAvailable = XQueryPointer(pPlatform->display, pPlatform->window, &root_return, &child_return, &root_x_ret, &root_y_ret, &win_x_ret, &win_y_ret, &mask_return);

    if (mouseAvailable) {
        // TODO: normalize coordinates from displaySpace to windowSpace
        // TODO: fire mouse movement event
        Vector2 m_coords = { (float) win_x_ret, (float) win_y_ret };
        return m_coords;
    }
    return Vector2::Zero;
}


Vector2 RWindowImpl::GetSize() const {
    return {(float) this->width, (float) this->height};
}

//Vector2 RWindow::getLastKnownResize() const
//{
//    return lastKnownWindowSize;
//}

// TODO: implement integer vector2/3 types
Vector2 RWindowImpl::GetPos() const {
    return pPlatform->position;
}

void RWindowImpl::SetPos(int x, int y) {
    XMoveWindow(pPlatform->display, pPlatform->window, x, y);
    pPlatform->position = { (float) x, (float) y };
}

void RWindowImpl::SetPos(const Vector2& pos) {
    SetPos(pos.x, pos.y);
}


void RWindowImpl::GLSwapBuffers() {
    glXSwapBuffers(pPlatform->display,pPlatform->window);
}


void RWindowImpl::Fullscreen() {
    Logger::Info(std::format("Fullscreening '{}'", this->title));
    fullscreen_mode = true;

    XEvent xev;
    Atom wm_state   = XInternAtom(pPlatform->display, "_NET_WM_STATE", true);
    Atom wm_fullscreen = XInternAtom(pPlatform->display, "_NET_WM_STATE_FULLSCREEN", true);

    XChangeProperty(pPlatform->display, pPlatform->window, wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)&wm_fullscreen, 1);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = pPlatform->window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 1; // _NET_WM_STATE_ADD
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(pPlatform->display, DefaultRootWindow(pPlatform->display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Fullscreened '{}'", this->title));
}

void RWindowImpl::RestoreFromFullscreen() {
    Logger::Debug(std::format("Restoring '{}' from Fullscreen", this->title));
    fullscreen_mode = false;
    XEvent xev;
    Atom wm_state = XInternAtom(pPlatform->display, "_NET_WM_STATE", False);
    Atom fullscreen = XInternAtom(pPlatform->display, "_NET_WM_STATE_FULLSCREEN", False);
    memset(&xev, 0, sizeof(xev));
    xev.type = ClientMessage;
    xev.xclient.window = pPlatform->window;
    xev.xclient.message_type = wm_state;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = 0; // _NET_WM_STATE_REMOVE
    xev.xclient.data.l[1] = fullscreen_mode;
    xev.xclient.data.l[2] = 0;
    XSendEvent(pPlatform->display, DefaultRootWindow(pPlatform->display), False, SubstructureNotifyMask | SubstructureRedirectMask, &xev);
    Logger::Debug(std::format("Restored '{}' from Fullscreen", this->title));
}

void RWindowImpl::SetVsyncEnabled(bool b) {
    PFNGLXSWAPINTERVALEXTPROC glXSwapIntervalEXT;
    glXSwapIntervalEXT = (PFNGLXSWAPINTERVALEXTPROC)glXGetProcAddressARB((const GLubyte*)"glXSwapIntervalEXT");
    glXSwapIntervalEXT(pPlatform->display, pPlatform->window, b);
}


// I know this doesn't modify the class, but it indirectly modifies the window
// Making it const just seems deceptive.
void RWindowImpl::SetCursorStyle(CursorStyle style) const {
    u32 x11_cursor_resolved_enum = static_cast<u32>(style.X11Cursor);
    Cursor c = XCreateFontCursor(pPlatform->display, x11_cursor_resolved_enum);
    XDefineCursor(pPlatform->display, pPlatform->window, c);
}

void RWindowImpl::Open() {
    pPlatform->xSetWindowAttributes.border_pixel = BlackPixel(pPlatform->display, pPlatform->defaultScreen);
    pPlatform->xSetWindowAttributes.background_pixel = BlackPixel(pPlatform->display, pPlatform->defaultScreen);
    pPlatform->xSetWindowAttributes.override_redirect = True;
    pPlatform->xSetWindowAttributes.event_mask = ExposureMask;
    //SetVsyncEnabled(vsync);
    if (renderer == RenderingAPI::OPENGL) {
        GLint glAttributes[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
        pPlatform->visual = glXChooseVisual(pPlatform->display, pPlatform->defaultScreen, glAttributes);
        pPlatform->glContext = glXCreateContext(pPlatform->display, pPlatform->visual, nullptr, GL_TRUE);
    }

    pPlatform->xSetWindowAttributes.colormap = XCreateColormap(pPlatform->display, RootWindow(pPlatform->display, pPlatform->defaultScreen), pPlatform->visual->visual, AllocNone);

    pPlatform->window = XCreateWindow(pPlatform->display, RootWindow(pPlatform->display, pPlatform->defaultScreen), 0, 0, width, height, 0, pPlatform->visual->depth,
                           InputOutput, pPlatform->visual->visual, CWBackPixel | CWColormap | CWBorderPixel | NoEventMask,
                           &pPlatform->xSetWindowAttributes);
    // Set window to floating because fucking tiling WMs
    pPlatform->windowTypeAtom = XInternAtom(pPlatform->display, "_NET_WM_WINDOW_TYPE", False);
    pPlatform->windowTypeUtilityAtom = XInternAtom(pPlatform->display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(pPlatform->display, pPlatform->window, pPlatform->windowTypeAtom, XA_ATOM, 32, PropModeReplace,
                    (unsigned char *)&pPlatform->windowTypeUtilityAtom, 1);
    //
    XSelectInput(pPlatform->display, pPlatform->window,
                 ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
                 PointerMotionMask |
                 PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask |
                 SubstructureNotifyMask | CWColormap );
    XMapWindow(pPlatform->display, pPlatform->window);
    XStoreName(pPlatform->display, pPlatform->window, title.c_str());
    pPlatform->wmDeleteWindow = XInternAtom(pPlatform->display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(pPlatform->display, pPlatform->window, &pPlatform->wmDeleteWindow, 1);

    if (renderer == RenderingAPI::OPENGL)
        glXMakeCurrent(pPlatform->display, pPlatform->window, pPlatform->glContext);

    // Get the position of the renderable area relative to the rest of the window.
    XGetWindowAttributes(pPlatform->display, pPlatform->window, &pPlatform->windowAttributes);
    pPlatform->render_area_position =  { (float) pPlatform->windowAttributes.x, (float) pPlatform->windowAttributes.y };

    open = true;

    processOnOpen();
}

void RWindowImpl::SetTitle(const std::string &title) {
    this->title = title;
    XStoreName(pPlatform->display, pPlatform->window, title.c_str());
}



Vector2 RWindowImpl::GetPositionOfRenderableArea() const {
    return pPlatform->render_area_position;
}

std::string RWindowImpl::getGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(glGetString(GL_VENDOR)));
}


// TODO: Implement ControllerButton map

