#include <rewindow/types/window.h>
#include "rewindow/logger/logger.h"
/*
std::string RWindowFlagToStr(RWindowFlags flag) {
    switch (flag) {
        case RWindowFlags::IN_FOCUS: return "IN_FOCUS";
        case RWindowFlags::FULLSCREEN: return "FULLSCREEN";
        case RWindowFlags::RESIZABLE: return "RESIZEABLE";
        case RWindowFlags::VSYNC: return "VSYNC";
        case RWindowFlags::QUIT: return "QUIT";
        case RWindowFlags::MAX_FLAG: return "MAX_FLAG";
        default:
            return "unimplemented flag";
    }
};
 */

std::string RWindowFlagToStr(RWindowFlags flag) {
    switch (flag) {
        case RWindowFlags::IN_FOCUS: return "IN_FOCUS";
        case RWindowFlags::FULLSCREEN: return "FULLSCREEN";
        case RWindowFlags::RESIZABLE: return "RESIZEABLE";
        case RWindowFlags::VSYNC: return "VSYNC";
        case RWindowFlags::QUIT: return "QUIT";
        case RWindowFlags::MAX_FLAG: return "MAX_FLAG";
        default:
            return "unimplemented flag";
    }
};


using namespace ReWindow;



/*
RWindow::RWindow(const std::string& wTitle, int wWidth, int wHeight, RenderingAPI wRenderer, bool wFullscreen, bool wResizable, bool wVsync)
: title(wTitle), width(wWidth), height(wHeight), renderer(wRenderer), fullscreen_mode(wFullscreen), resizable(wResizable), vsync(wVsync),
flags{false,wFullscreen,wResizable,wVsync} {

}
 */


//RWindow::~RWindow() {
    /*
    if (open)
        DestroyOSWindowHandle();
    */
   //RWindowImpl::~RWindowImpl();
//}



Vector2 RWindowImpl::GetMouseCoordinates() const {
    return currentMouse.Position;
}

bool RWindowImpl::GetFlag(RWindowFlags flag) const {
    return flags[(int) flag];
}

bool RWindowImpl::IsAlive() const {
    return (!closing) && open;
}

void RWindowImpl::SetFullscreen(bool fs) {
    if (fs)
        Fullscreen();
    else
        RestoreFromFullscreen();
}

#pragma region Event Processors Implementation
void RWindowImpl::processFocusIn()
{
    RWindowEvent event {};
    OnFocusGain(event);
    OnFocusGainEvent(event);
    LogEvent(event);
}

void RWindowImpl::processFocusOut()
{
    RWindowEvent event {};
    OnFocusLost(event);
    OnFocusLostEvent(event);
    LogEvent(event);
}

void RWindowImpl::processMousePress(const MouseButton& btn)
{
    currentMouse.Set(btn, true);
    auto event = MouseButtonDownEvent(btn);
    OnMouseButtonDown(event);
    OnMouseButtonDownEvent(event);
    LogEvent(event);

}

void RWindowImpl::processMouseMove(Vector2 last_pos, Vector2 new_pos)
{
    currentMouse.Position = new_pos;
    auto event = MouseMoveEvent(new_pos);
    OnMouseMove(event);
    OnMouseMoveEvent(event);
    LogEvent(event);
}

void RWindowImpl::processMouseRelease(const MouseButton& btn)
{
    currentMouse.Set(btn, false);
    auto event = MouseButtonUpEvent(btn);
    OnMouseButtonUp(event);
    OnMouseButtonUpEvent(event);
    LogEvent(event);

}

void RWindowImpl::processKeyRelease(Key key) {
    currentKeyboard.PressedKeys[key] = false;
    auto event = KeyUpEvent(key);
    OnKeyUp(event);
    OnKeyUpEvent(key);
    LogEvent(event);
}

void RWindowImpl::processKeyPress(Key key) {
    currentKeyboard.PressedKeys[key] = true;
    auto event = KeyDownEvent(key);
    OnKeyDown(event);
    OnKeyDownEvent(key);
    LogEvent(event);
}

void RWindowImpl::processOnClose()
{
    auto event = RWindowEvent();
    OnClosing();
    OnClosingEvent();
    LogEvent(event);
}

void RWindowImpl::processOnOpen()
{
    auto event = RWindowEvent();
    OnOpen();
    OnOpenEvent();
    LogEvent(event);
}

#pragma endregion

std::string RWindowImpl::GetTitle() const {
    return this->title;
}

/*
Vector2 RWindowImpl::GetSize() const
{
    return {this->width, this->height};
}
*/

int RWindowImpl::GetWidth() const
{
    return this->width;
}

int RWindowImpl::GetHeight() const
{
    return this->height;
}

void RWindowImpl::SetSizeWithoutEvent(const Vector2& size) {
    width = size.x;
    height = size.y;
}

void RWindowImpl::SetLastKnownWindowSize(const Vector2& size) {
    lastKnownWindowSize = size;
}

void RWindowImpl::SetRenderer(RenderingAPI api) {
    renderer = api;
}

void RWindowImpl::SetSize(const Vector2& size) {
    this->width = size.x;
    this->height = size.y;
    this->SetSize(size.x, size.y);
}

bool RWindowImpl::IsKeyDown(Key key) const {
    if (currentKeyboard.PressedKeys.contains(key))
        return currentKeyboard.PressedKeys.at(key);

    return false; // NOTE: Key may not be mapped!!
}

bool RWindowImpl::IsMouseButtonDown(const MouseButton &button) const {
    // TODO: Implement MouseButton map
    return currentMouse.IsDown(button);
}


void RWindowImpl::ManagedRefresh()
{
     auto begin = GetTimestamp();
     Refresh();
     auto end = GetTimestamp();

     float dt = ComputeElapsedFrameTimeSeconds(begin, end);

     UpdateFrameTiming(dt);
}

/*
void RWindowImpl::Refresh() {
    PollEvents();
    OnRefresh(delta_time);

    // Only call once and cache the result.
    currentMouse.Position = GetAccurateMouseCoordinates();

    /// TODO: Implement optional minimum epsilon to trigger a Mouse Update.
    if (currentMouse.Position != previousMouse.Position) {
        processMouseMove(previousMouse.Position, currentMouse.Position);
        previousMouse.Position = currentMouse.Position;
    }
}
 */

float RWindowImpl::ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end) {
    auto frame_time = end - start;
    unsigned long int frame_time_us = std::chrono::duration_cast<std::chrono::microseconds>(frame_time).count();
    float frame_time_s = frame_time_us / (1000.f * 1000.f);
    return frame_time_s;
}

 std::chrono::steady_clock::time_point RWindowImpl::GetTimestamp() {
    return std::chrono::steady_clock::now();
}

void RWindowImpl::UpdateFrameTiming(float frame_time) {
    delta_time = frame_time;
    refresh_rate = 1.f / delta_time;
    refresh_rate_prev_5 = refresh_rate_prev_4;
    refresh_rate_prev_4 = refresh_rate_prev_3;
    refresh_rate_prev_3 = refresh_rate_prev_2;
    refresh_rate_prev_2 = refresh_rate_prev_1;
    refresh_rate_prev_1 = refresh_rate;
    avg_refresh_rate = (refresh_rate_prev_1 + refresh_rate_prev_2 + refresh_rate_prev_3 + refresh_rate_prev_4 + refresh_rate_prev_5) / 5.f;
    refresh_count++;
}


void RWindowImpl::processMouseWheel(int scrolls)
{
    currentMouse.Wheel += scrolls;
    auto ev = MouseWheelEvent(scrolls);
    OnMouseWheel(ev);
    OnMouseWheelEvent(ev);

    previousMouse.Wheel = currentMouse.Wheel;
}


bool RWindowImpl::IsResizable() const {
    return resizable;
}

bool RWindowImpl::IsFullscreen() const {
    return fullscreen_mode;
}

bool RWindowImpl::IsFocused() const {
    return focused;
}

 bool RWindowImpl::IsVsyncEnabled() const {
     return vsync;
 }

 float RWindowImpl::GetDeltaTime() const { return delta_time; }

 float RWindowImpl::GetRefreshRate() const { return refresh_rate; }

 float RWindowImpl::GetRefreshCounter() const { return refresh_count; }

 void RWindowImpl::Close() {
     closing = true;
     processOnClose();
 }

 void RWindowImpl::ForceClose() {
     Close();
     DestroyOSWindowHandle();
 }

 void RWindowImpl::ForceCloseAndTerminateProgram() {
     ForceClose();
     exit(0);
 }



