#include <Windows.h>
#include <gl/GL.h>
#include <rewindow/types/window.h>

using namespace ReWindow;

bool fullscreenmode = false;
bool open = false;
HINSTANCE hInstance = GetModuleHandle(nullptr);
HWND hwnd;
HDC hdc;
HGLRC glContext;

void raise() { SetWindowPos(hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }
void lower() { SetWindowPos(hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }

void RWindow::SetFlag(RWindowFlags flag, bool state) {
    flags[(int) flag] = state;
    if (flag == RWindowFlags::RESIZABLE && !state) {
        RECT rect;
        GetWindowRect(hwnd, &rect);
        LONG style = GetWindowLong(hwnd, GWL_STYLE);
        style &= ~(WS_THICKFRAME | WS_MAXIMIZEBOX);
        SetWindowLong(hwnd, GWL_STYLE, style);
        SetWindowPos(hwnd, nullptr, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER);
    }
}wasd
void RWindow::SetResizable(bool resizable) {
    SetFlag(RWindowFlags::RESIZABLE, resizable);;
}


void RWindow::PollEvents() {
    MSG msg;
    while (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}

void RWindow::SetSize(int newWidth, int newHeight) {
    if (!resizable) return;
    this->width = width;
    this->height = height;
    SetWindowPos(hwnd, nullptr, 0, 0, width, height, SWP_NOMOVE | SWP_NOZORDER);
}

Vector2 RWindow::GetAccurateMouseCoordinates() const {
    POINT point;
    GetCursorPos(&point);
    ScreenToClient(hwnd, &point);
    return { (float)point.x, (float)point.y };
}

void RWindow::SetCursorVisible(bool cursor_enable) {
    cursor_visible = cursor_enable;
}

bool RWindow::GetCursorVisible() {
    return cursor_visible;
}


Vector2 RWindow::GetSize() const {
    RECT rect;
    GetClientRect(hwnd, &rect);
    return { (float)(rect.right - rect.left), (float)(rect.bottom - rect.top) };
}

void RWindow::SetPos(int x, int y) {
    SetWindowPos(hwnd, nullptr, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

void RWindow::SetPos(const Vector2& pos) {
    SetPos(pos.x, pos.y);
}

void RWindow::Fullscreen() {
    // Implement fullscreen
    SetWindowLong(hwnd, GWL_STYLE, GetWindowLong(hwnd, GWL_STYLE) & ~WS_OVERLAPPEDWINDOW);
    SetWindowPos(hwnd, HWND_TOP, 0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN), SWP_FRAMECHANGED | SWP_NOOWNERZORDER);
}

void RWindow::RestoreFromFullscreen() {
    // Implement restore from fullscreen
    SetWindowLong(hwnd, GWL_STYLE, GetWindowLong(hwnd, GWL_STYLE) | WS_OVERLAPPEDWINDOW);
    SetWindowPos(hwnd, nullptr, 0, 0, width, height, SWP_FRAMECHANGED | SWP_NOZORDER | SWP_NOOWNERZORDER);
}

void RWindow::SetVsyncEnabled(bool b) {
    typedef BOOL(WINAPI* PFNWGLSWAPINTERVALEXTPROC)(int interval);
    auto wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");

    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(b ? 1 : 0);
}

RWindow* eWindow = nullptr;
KeyboardState* pKeyboard = nullptr;
KeyboardState* cKeyboard = nullptr;

//Event loop.
LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    switch (uMsg) {
            case WM_CLOSE: {
                    DestroyWindow(hwnd);

                    eWindow->processOnClose();
            }

            case WM_DESTROY: {
                    exit(0);
            }

            case WM_SIZE: {
                    eWindow->SetSizeWithoutEvent({(float) LOWORD(lParam), (float) HIWORD(lParam)});
                    auto eventData = WindowResizeRequestEvent();
                    eventData.Size = {(float) eWindow->GetWidth(), (float) eWindow->GetHeight()};
                    eWindow->SetLastKnownWindowSize({(float) eWindow->GetWidth(), (float) eWindow->GetHeight()});

                    // TODO: Implement eWindow->processOnResize()

                    eWindow->OnResizeRequest(eventData);
                    eWindow->OnResizeRequestEvent(eventData);


                    //Just to be absolutely sure the OpenGL viewport resizes along with the window.
                    glViewport(0, 0, eWindow->GetWidth(), eWindow->GetHeight());
                    break;
                }

            case WM_SETFOCUS: {

                    eWindow->processFocusIn();

                    eWindow->SetFlag(RWindowFlags::IN_FOCUS, true);
                    break;
            }

            case WM_KILLFOCUS: {

                    eWindow->processFocusOut();

                    eWindow->SetFlag(RWindowFlags::IN_FOCUS, false);
                    break;
            }

            case WM_SETCURSOR: {
                    if (LOWORD(lParam) == HTCLIENT && eWindow->GetCursorVisible() == false)
                        SetCursor(nullptr);
                break;
            }

            case WM_KEYDOWN: {
                    auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);
                    //Key repeat fix.
                    if (!pKeyboard->PressedKeys[key]) {
                        eWindow->processKeyPress(key);
                    }
                    break;
            }

            case WM_KEYUP: {
                    auto key = GetKeyFromWindowsScancode((WindowsScancode) wParam);
                    eWindow->processKeyRelease(key);
                    break;
            }

            //Mouse Buttons.
            case WM_MOUSEWHEEL: {
                    int wheel_delta = GET_WHEEL_DELTA_WPARAM(wParam);

                    // TODO: Determine sign of wheel_delta for each direction, (and on linux too), and document this.
                    eWindow->processMouseWheel(wheel_delta);

                    break;
            }

            case WM_LBUTTONDOWN: {
                    eWindow->processMousePress(MouseButtons::Left);
                    break;
            }

            case WM_LBUTTONUP: {
                    eWindow->processMouseRelease(MouseButtons::Left);
                    break;
            }

            case WM_RBUTTONDOWN: {
                    eWindow->processMousePress(MouseButtons::Right);
                    break;
            }

            case WM_RBUTTONUP: {
                    eWindow->processMouseRelease(MouseButtons::Right);
                    break;
            }

            case WM_MBUTTONDOWN: {
                    eWindow->processMousePress(MouseButtons::Middle);
                    break;
            }

            case WM_MBUTTONUP: {
                    eWindow->processMouseRelease(MouseButtons::Middle);
                    break;
            }

            case WM_XBUTTONDOWN: {
                    WORD button = GET_XBUTTON_WPARAM(wParam);

                    if (button == XBUTTON1)
                        eWindow->processMousePress(MouseButtons::Mouse4);

                    if (button == XBUTTON2)
                        eWindow->processMousePress(MouseButtons::Mouse5);

                    break;
            }

            case WM_XBUTTONUP: {
                    WORD button = GET_XBUTTON_WPARAM(wParam);
                    if (button == XBUTTON1)
                        eWindow->processMouseRelease(MouseButtons::Mouse4);

                    if (button == XBUTTON2)
                        eWindow->processMouseRelease(MouseButtons::Mouse5);
                    break;
        }

        //This is the same as "Motion Notify" in the X Window System.
        case WM_MOUSEMOVE:
            break;
    }

    if (pKeyboard != nullptr && cKeyboard != nullptr)
        pKeyboard = cKeyboard;
    
    return DefWindowProc(hwnd, uMsg, wParam, lParam);
}

void RWindow::SetCursorStyle(CursorStyle style) const {}

void RWindow::Open() {
    eWindow = this;
    pKeyboard = &previousKeyboard;
    cKeyboard = &currentKeyboard;
    WNDCLASS wc = { };
    wc.lpfnWndProc = WindowProc;
    wc.hInstance = hInstance;
    wc.lpszClassName = "RWindowClass";
    RegisterClass(&wc);
    hwnd = CreateWindowEx(
        0,
        "RWindowClass",
        title.c_str(),
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT, width, height,
        nullptr,
        nullptr,
        hInstance,
        nullptr
    );

    if (renderer == RenderingAPI::OPENGL) {
        PIXELFORMATDESCRIPTOR pfd = {
            sizeof(PIXELFORMATDESCRIPTOR),
            1,
            PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
            PFD_TYPE_RGBA,
            24,
            0, 0, 0, 0, 0, 0,
            0,
            0,
            0,
            0, 0, 0, 0,
            24,
            8,
            0,
            PFD_MAIN_PLANE,
            0,
            0, 0, 0
        };

        hdc = GetDC(hwnd);
        int pixelFormat = ChoosePixelFormat(hdc, &pfd);
        SetPixelFormat(hdc, pixelFormat, &pfd);

        glContext = wglCreateContext(hdc);
        wglMakeCurrent(hdc, glContext);
    }
    ShowWindow(hwnd, SW_SHOW);
    open = true;
}

void RWindow::GLSwapBuffers() {
    SwapBuffers(hdc);
}

std::string RWindow::getGraphicsDriverVendor() {
    return std::string(reinterpret_cast<const char*>(glGetString(GL_VENDOR)));
}
